module Graph where

import Data.List

{-
    Urmatoarele exercitii vor avea drept scop implementarea unei mici biblioteci
    pentru grafuri ORIENTATE.
    Dupa cum stiti, exista mai multe modalitati de reprezentare a unui graf.
    Biblioteca noastra va defini mai multe astfel de reprezentari, precum
    si algoritmi care opereaza pe grafuri.
    Algoritmii vor fi independenti de reprezentarea interna a grafului - ei
    vor functiona indiferent de ce structura de date am ales noi pentru un anume
    graf.
    Pentru a obtine aceasta genericitate vom abstractiza notiunea de graf intr-o
    clasa care va expune operatiile pe care orice structura de date de tip graf
    ar trebui sa le aiba.
-}

-- reprezentam nodurile ca intregi
type Node = Int

-- reprezentam arcele ca perechi de noduri
type Arc = (Node, Node) 

-------------------------------------------------------------------------------

{-
    1.(1.5p) Analizati clasa Graph definita mai jos si scrieti implementarile
    default pentru functiile din aceasta clasa.
    
    Clasa Graph care defineste interfata pentru toate structurile de grafuri
    pe care le vom defini mai jos.
-}
class Graph g where
    
    -- Construieste un graf plecand de la o lista de noduri si arcele
    -- dintre noduri
    build :: [Node] -> [Arc] -> g

    -- Lista tuturor nodurilor din graf
    nodes :: g -> [Node] -- lista nodurilor din graf
    
    -- Lista arcelor din graf
    arcs :: g -> [Arc] -- lista muchiilor din graf

    -- Lista nodurilor catre care nodul dat ca parametru are un arc
    nodeOut :: g -> Node -> [Node]

    -- lista nodurilor care au un arc catre nodul dat ca parametru
    nodeIn :: g -> Node -> [Node]
    
    -- verifica faca exista un arc intre 2 noduri
    arcExists :: g -> Node -> Node -> Bool

    -- TODO: implementari default
    arcExists g a b = (a, b) `elem` arcs g
    arcs g = [(i, j) | i <- ns, j <- ns, arcExists g i j] where ns = nodes g
    nodeIn g n = [x | x <- nodes g, arcExists g x n]
    nodeOut g n = [x | x <- nodes g, arcExists g n x]

-------------------------------------------------------------------------------
{-
    2.(2p) Definiti tipul AdjListGraph care reprezinta un graf ca pe o serie de
    perechi (nod, lista vecini). Includeti AdjListGraph in clasa Graph
-}

newtype AdjListGraph = ALGraph [(Node, [Node])] deriving Show

instance Graph AdjListGraph where
    build ns arcs = ALGraph [(node, adjacentNodes node) | node <- ns]
        where adjacentNodes node = [neighbour | (n, neighbour) <- arcs, n == node]

    nodes (ALGraph ps) = map fst ps

    arcs (ALGraph ps) = [(node, neighbour) | (node, neighbours) <- ps, neighbour <- neighbours]

    nodeOut (ALGraph ps) n = let Just neighbours = lookup n ps in neighbours


-------------------------------------------------------------------------------

{-
    3. (1p) Definiti tipul ArcGraph care reprezinta un graf ca o lista de noduri
    si o lista de arce dintre noduri. Includeti ArcGraph in clasa Graph
-}


data ArcGraph = AGraph [Node] [Arc] deriving (Show)

instance Graph ArcGraph where
    build ns arcs = AGraph ns arcs

    nodes (AGraph ns _) = ns

    arcs (AGraph _ as) = as

-------------------------------------------------------------------------------

{-
    4. (0.5p) Definiti functia convert care face conversia intre reprezentari diferite
    de grafuri.    
-}
convert :: (Graph g1, Graph g2) => g1 -> g2
convert g1 = build (nodes g1) (arcs g1)


-------------------------------------------------------------------------------
{-
    O traversare a unui graf este o lista de perechi (nod, Maybe parinte-node). Aceasta
    structura va contine toate nodurile rezultate in urma unei parcurgeri a unui graf,
    in ordinea aparitiei in parcurgere si impreuna cu parintele nodului din parcurgere
    (Pentru un nod N, parintele sau este nodul din care s-a ajuns la N in decursul parcurgerii)
-}
type Traversal = [(Node, Maybe Node)]

-- Un o cale in graf este reprezentata ca lista de noduri din acea cale.
type Path = [Node]

{-
    Definitie pentru algoritmi de parcurgere a unui graf. Un algoritm de parcurgere este o functie
    care primeste un graf si un nod de start si intoarce o structura Traversal. Observati
    ca tipul grafului este o variabila de tip - algoritmii trebuie sa functioneze pentru orice
    structura de graf.
-}
type TraverseAlgo g = g -> Node -> Traversal

{-
    5. (2p) Implementati algoritmul de parcurgere in adancime. (Depth-First Search)
-}
dfs :: Graph g => TraverseAlgo g
dfs = mkTraversal (++)

{-
    6. (2p) Implementati algoritmul de parcurgere in latime. (Breadth-First Search)
-}
bfs :: Graph g => TraverseAlgo g
bfs = mkTraversal (flip (++))

-- DFS si BFS descriu in principiu acelasi proces - folosesc o structura de date in
-- care sunt tinute nodurile care trebuie parcurse in continuare, precum si
-- o structura cu nodurile vizitate deja. Ceea ce difera este structura de
-- date (DFS - stiva, BFS - coada). Vom folosi o lista pentru a repr. ambele
-- structuri de date - si deci ce variaza este modul in care adaugam nodurile
-- de parcurs - la DFS adaugam la inceputul listei, la BFS - la sfarsit.
mkTraversal accumOp g start = iter [(start, Nothing)] []
    where
        iter [] visited = reverse visited
        iter (p@(x, parent):xs) visited
            | any ((==x).fst) visited  = iter xs visited
            | otherwise =
                let additions = [(y, Just x) | y <- nodeOut g x]
                in  iter (accumOp additions xs) (p:visited)


-------------------------------------------------------------------------------

{-
    7. (1p) Definiti functia findPath care primeste ca parametri:
    * un algoritmi de parcurgere
    * un graf
    * un nod de pornire
    * un nod destinatie
    si care intoarce o cale dintre cele 2 noduri, daca o astfel de cale
    exista.
-}
findPath :: Graph g => TraverseAlgo g -> g -> Node -> Node -> Maybe Path
findPath traverse g n1 n2 = fmap reverse (connect n1 n2)
    where
        traversal = traverse g n1
        connect n1 n2
            | n1 == n2 = Just [n1]
            | otherwise = do
                Just parent <- lookup n2 traversal
                subPath     <- connect n1 parent
                return (n2 : subPath)

-------------------------------------------------------------------------------

{-
    8. (BONUS 0.5p). Creati tipul GenericAdjListGraph care este este similar cu
    AdjListGraph, dar in care nodurile nu sunt neaparat de tip Int - pot
    fi de orice tip.
-}

newtype GenericAdjListGraph a = GALGraph [(a, [a])] deriving Show


{-
    9. (BONUS 0.5p). Adaugati tipul GenericAdjListGraph la clasa Functor
-}
instance Functor GenericAdjListGraph where
    fmap f (GALGraph ps) = GALGraph [(f x, map f xs) | (x, xs) <- ps] 

{-
    10. (BONUS 1p). Definiti clasa Listable pentru containere care contin elemente
    si care va avea o singura functie: toList - care intoarce lista elementelor
    continute in container. Adaugati GenericAdjListGraph la aceasta clasa.
-}

class Listable c where
    toList :: c a -> [a]

instance Listable GenericAdjListGraph where
    toList (GALGraph ps) = map fst ps

